#include <openssl/rand.h>
#include <openssl/pem.h>
#include <openssl/ssl.h>
#include <openssl/rsa.h>
#include <openssl/evp.h>
#include <openssl/bio.h>
#include <openssl/evp.h>
#include <openssl/err.h>
#include <openssl/bn.h>

#include <stdio.h>
#include <string.h>
#include <stdbool.h>

#include "smol_rc4.h"

int RSA_PADDING = RSA_PKCS1_PADDING;

bool check_pub_keys(const RSA* rsa1, const RSA* rsa2) {
    if(rsa1 == NULL || rsa2 == NULL) {
        return false;
    }
    const BIGNUM *n1 = RSA_get0_n(rsa1);
    const BIGNUM *n2 = RSA_get0_n(rsa2);
    const BIGNUM *e1 = RSA_get0_e(rsa1);
    const BIGNUM *e2 = RSA_get0_e(rsa2);

    if(BN_cmp(n1, n2) != 0) return false;
    if(BN_cmp(e1, e2) != 0) return false;

    return true;
}


RSA * createRSA(unsigned char * key, int public) {
    RSA *rsa = NULL;
    BIO *keybio ;
    keybio = BIO_new_mem_buf(key, -1);
    if (keybio==NULL) {
        printf( "!ERR!: Failed to create key BIO");
        return NULL;
    }
    if(public) rsa = PEM_read_bio_RSA_PUBKEY(keybio, &rsa,NULL, NULL);
    else rsa = PEM_read_bio_RSAPrivateKey(keybio, &rsa,NULL, NULL);

    if(rsa == NULL) {
        printf( "!ERR!: Failed to create RSA");
    }
 
    return rsa;
}

int public_decrypt(RSA* rsa, unsigned char * enc_data,int data_len, unsigned char *decrypted) {
    return RSA_public_decrypt(data_len,enc_data,decrypted,rsa,RSA_PADDING);
}


int public_encrypt(RSA* rsa, unsigned char * data,int data_len, unsigned char *encrypted) {
    return RSA_public_encrypt(data_len,data,encrypted,rsa,RSA_PADDING);
}


int private_decrypt(RSA * rsa, unsigned char * enc_data,int data_len, unsigned char *decrypted) {
    return RSA_private_decrypt(data_len,enc_data,decrypted,rsa,RSA_PADDING);
}



int private_encrypt(RSA * rsa, unsigned char * data,int data_len, unsigned char *encrypted) {
    return RSA_private_encrypt(data_len,data,encrypted,rsa,RSA_PADDING);
}



void printLastError(char *msg)
{
    char * err = malloc(130);;
    ERR_load_crypto_strings();
    ERR_error_string(ERR_get_error(), err);
    printf("%s ERROR: %s\n",msg, err);
    free(err);
}


char* read_file(const char* filename){
    char * buffer = 0;
    long length;
    FILE * fhandle = fopen (filename, "rb");

    if (fhandle) {
      fseek(fhandle, 0, SEEK_END);
      length = ftell(fhandle);
      fseek(fhandle, 0, SEEK_SET);
      buffer = calloc (length, sizeof(char));
      if (buffer) {
        fread (buffer, 1, length, fhandle);
      }
      fclose (fhandle);
    }

    return buffer;
}


unsigned char decode(x) char x; {
	if (x >= '0' && x <= '9') {
        return (x - 0x30);
    }
    else if (x >= 'A' && x <= 'F') { 
        return(x - 0x37);
    }
    else if (x >= 'a' && x <= 'f') {
        return(x - 0x57);
    }
    else {
    	fprintf(stderr,"\nInput is not in legal hex format\n");
        exit(1);
    }
}


unsigned char* unhex(const char* hexstring) {
    size_t len = strlen(hexstring);
    if (len % 2 != 0) return NULL;

    unsigned char* result = calloc(len/2 + 1, sizeof(unsigned char));

    for(int i=0; i<len; i+=2) {
        char l = hexstring[i];
        char r = hexstring[i+1];
        result[i/2] = ((decode(l) << 4) & 0xf0) + (decode(r) & 0x0f);
    }
    result[len/2] = '\0';
    return result;
}


ssize_t readline(char** line, size_t* len) {
	ssize_t read = -1;
    if ((read = getline(line, len, stdin)) != -1) {
        if (read>0 && (*line)[read -1] == '\n') {
            (*line)[read-1] = '\0';
            read -= 1;
        }
    } else {
        printf("!ERR!: FAILED TO READ DATA");        
        return -1;
    }
    return read;
}


int main() {
    // Initialize my keys
    unsigned char* privateKey = (unsigned char*)read_file("private.pem");
    unsigned char* publicKey = (unsigned char*)read_file("public.pem");

    unsigned char* ciphertext = NULL;

    unsigned char *decrypted = malloc(65536);
    unsigned char *encrypted = malloc(65536);
    int decryptedLength = 0;
    int encryptedLength = 0;
    size_t rc4_key_length = 16;
    unsigned char *rc4_key_client = malloc(rc4_key_length);
    unsigned char *rc4_key_server = malloc(rc4_key_length);
  
    RSA *rsaPub_client = NULL;
	RSA *rsaPub = createRSA(publicKey, 1);
    RSA *rsaPriv = createRSA(privateKey, 0);
    free(privateKey);
	privateKey = NULL;
	free(publicKey);
	publicKey = NULL;


    if(rsaPub == NULL) {
        printf( "!ERR!: Failed to create PUB RSA");
        return -1;
    }

    if(rsaPriv == NULL) {
        printf( "!ERR!: Failed to create PRIV RSA");
        return -1;
    }

    char* line = calloc(65536*2 + 1, sizeof(char));
    size_t len = 65536*2;
    ssize_t read = 0;

    read = readline(&line, &len);

    if (read % 2 != 0) {
        printf("!ERR!: HEX STRING MUST HAVE EVEN LENGTH");
        return -1;
    }

    // Decode client RC4 key
    ciphertext = unhex(line);

    for(int i=0; i<(read / 2); i++) {
        printf("%02X", ciphertext[i]);
    }
    printf("\n");

    encryptedLength = read / 2;

    // Decrypt client RC4 key
    decryptedLength = private_decrypt(rsaPriv, ciphertext, encryptedLength, decrypted);
    if(decryptedLength == -1) {
        printLastError("!ERR!: Private Key Decrypt failed ");
        return -1;
    }

    memcpy(rc4_key_client, decrypted, rc4_key_length);
    for(int i=0; i<rc4_key_length; i++) {
        printf("%02X", rc4_key_client[i]);
    }
    printf("\n");

    // Check key length
    if(decryptedLength != rc4_key_length) {
        printf("!ERR!: EXPECTED 16 RC4 CLIENT KEY");
        return -1;        
    }

    // Generate server RC4 Key
    if (RAND_bytes(rc4_key_server, rc4_key_length) != 1){
        printf("!ERR!: Failed to generate random key");
        return -1;
    }

    // Read client auth name
    read = readline(&line, &len);

    // Decode client auth name
    unsigned char* name = unhex(line);
    size_t nameLength = read/2;

    decryptedLength = RC4(rc4_key_client, name, nameLength, decrypted);
    if(decryptedLength == 0) {
        printf("!ERR!: RC4 DECRYPTION FAILED FOR AUTH NAME");
        return -1;
    }

    memcpy(name, decrypted, decryptedLength);
    if(strlen((char*)name) > 100) {
        printf("!ERR!: AUTH NAME TOO LONG... %lu > 100", strlen((char*)name));
        return -1;
    }

    // path = "./$name/rsa.pub" -> 2 + name + 8 + NULLbyte
    char* auth_key_path = calloc(strlen((char*)name) + 11, sizeof(char));
    strcpy(auth_key_path, "./");
    strcpy(auth_key_path + 2, (const char*)name);
    strcpy(auth_key_path + (strlen((char*)name) + 2), "/rsa.pub");
    // check path
    for(int i=0; i<strlen(auth_key_path) -1; i++) {
        if(auth_key_path[i] == '.' && auth_key_path[i+1] == '.') {
            printf("!ERR!: ILLEGAL PATH DETECTED!");
            return -1;
        }
    }

    // Load know pubkey for client
    unsigned char* clientPubKey_known = (unsigned char*)read_file(auth_key_path);
    RSA * RSAclientPub_known = createRSA(clientPubKey_known, 1);

    // Read client pub key
    read = readline(&line, &len);

    // Decode client pub key
    ciphertext = unhex(line);
    encryptedLength = read/2;

    // int RC4(unsigned char *key, char *input, size_t inputLength, unsigned char *output);
    decryptedLength = RC4(rc4_key_client, ciphertext, encryptedLength, decrypted);
    if(decryptedLength == 0) {
        printf("!ERR!: RC4 DECRYPTION FAILED FOR CLIENT PUB KEY");
        return -1;
    }

    unsigned char* clientPublicKey_in = decrypted; 
    rsaPub_client = createRSA(clientPublicKey_in, 1);
    clientPublicKey_in = NULL;

    // Verify pub keys
    if(check_pub_keys(RSAclientPub_known, rsaPub_client) != true) {
        printf("!ERR!: RSA AUTH KEY VERIFICATION FAILED!");
        return -1;
    }

    // Encrypt server's RC4 key
    encryptedLength = public_encrypt(rsaPub_client, rc4_key_server, rc4_key_length, encrypted);
    if(encryptedLength == -1) {
        printLastError("Public Encrypt failed ");
        exit(0);
    }

    // Hex encode RC4 Key (and send)
    for(int i=0; i<rc4_key_length; i++) {
        printf("%02X", rc4_key_server[i]);
    }
    printf("\n");

    // Read secret for client
    // path = "./$name/flag" -> 2 + name + 5 + NULLbyte
    char* secretPath = calloc(strlen((char*)name) + 8, sizeof(char));
    strcpy(secretPath, "./");
    strcpy(secretPath + 2, (const char*)name);
    strcpy(secretPath + (strlen((char*)name) + 2), "/flag");
    // check path
    for(int i=0; i<strlen(auth_key_path) -1; i++) {
        if(auth_key_path[i] == '.' && auth_key_path[i+1] == '.') {
            printf("!ERR!: ILLEGAL PATH DETECTED!");
            return -1;
        }
    }
    unsigned char* secret = (unsigned char*)read_file(secretPath);

    // Results for: $name\n$secret
    size_t resultSize = 16 + nameLength + strlen((char*)secret);
    if(resultSize >= (65535)) {
            printf("!ERR!: OUTPUT TOO LONG!");
            return -1;
    }
    memcpy(decrypted, "Secret data for: ", 17);
    memcpy(decrypted, "Results for: ", 17);
    memcpy(decrypted + 17, name, nameLength);
    memcpy(decrypted + 17 + nameLength, "\n", 1);
    memcpy(decrypted + 17 + nameLength + 1, secret, strlen((char*)secret));

    decryptedLength = 17 + nameLength + 1 + strlen((char*)secret);

    // int RC4(unsigned char *key, char *input, size_t inputLength, unsigned char *output);
    encryptedLength = RC4(rc4_key_server, decrypted, decryptedLength, encrypted);
    if (decryptedLength != encryptedLength) {
            printf("!ERR!: RC4 ENCRYPTION EXPECTED RESULT SIZE MISSMATCH!");
            return -1;
    }

    // send encrypted secret for authenticated user
    for(int i=0; i<encryptedLength; i++) {
        printf("%02X", encrypted[i]);
    }
    printf("\n");

    // we should clean up, but whatever, just die.
    return 0;
}
